<?php

namespace common\logic\Document\Document;

use common\logic\BaseService;
use common\logic\Order\Order\Order;
use common\logic\Producer\Producer\Producer;
use common\logic\SolverInterface;

class DocumentSolver extends BaseService implements SolverInterface
{
    public function getAmount(DocumentInterface $document, $type = Order::AMOUNT_TOTAL, $format = false)
    {
        return $this->_getAmountGeneric($document, $type, false, $format);
    }

    public function getAmountWithTax(DocumentInterface $document, $type = Order::AMOUNT_TOTAL, $format = false)
    {
        return $this->_getAmountGeneric($document, $type, true, $format);
    }

    protected function _getAmountGeneric(DocumentInterface $document, $type = Order::AMOUNT_TOTAL, $withTax = true, $format = false)
    {
        $amount = 0;
        $totalVat = 0;

        foreach($document->getProductsOrders() as $productOrderArray) {
            foreach($productOrderArray as $productOrder) {
                $priceLine = $productOrder->getPriceByTypeTotal($type) * $productOrder->quantity;
                $amount += $priceLine;
                $totalVat += Price::getVat($priceLine, $productOrder->taxRate->value, $document->tax_calculation_method);
            }
        }

        if ($this->isTaxCalculationMethodRoundingOfTheSum()) {
            $totalVat = Price::round($totalVat);
        }

        if ($withTax) {
            $amount += $totalVat;
        }

        if ($format) {
            return Price::format($amount);
        } else {
            return $amount;
        }
    }

    public function getTotalVatArray($typeTotal)
    {
        $totalVatArray = [];

        // Méthode de calcul via getProductOrders()
        foreach($this->getProductsOrders() as $productOrderArray) {
            foreach ($productOrderArray as $productOrder) {

                $idTaxRate = $productOrder->taxRate->id;
                if (!isset($totalVatArray[$idTaxRate])) {
                    $totalVatArray[$idTaxRate] = 0;
                }

                $totalVatArray[$idTaxRate] += Price::getVat(
                    $productOrder->getPriceByTypeTotal($typeTotal) * $productOrder->quantity,
                    $productOrder->taxRate->value,
                    $this->tax_calculation_method
                );
            }
        }

        return $totalVatArray;
    }

    public function getPointSale()
    {
        if (isset($this->orders) && isset($this->orders[0])) {
            return $this->orders[0]->pointSale;
        } else {
            return '';
        }
    }

    public function getDistribution()
    {
        if (isset($this->orders) && isset($this->orders[0])) {
            return $this->orders[0]->distribution;
        } else {
            return '';
        }
    }

    public function getClass()
    {
        return str_replace('common\models\\', '', get_class($this));
    }

    public function getType()
    {
        $class = $this->getClass();

        if ($class == 'Invoice') {
            $documentType = 'Facture';
        } elseif ($class == 'DeliveryNote') {
            $documentType = 'Bon de livraison';
        } elseif ($class == 'Quotation') {
            $documentType = 'Devis';
        }

        if (isset($documentType)) {
            return $documentType;
        }

        return '';
    }

    public function isValidClass($typeDocument)
    {
        return in_array($typeDocument, ['Invoice', 'DeliveryNote', 'Quotation']);
    }

    public function getStatusWording(DocumentInterface $document): string
    {
        return ($document->status == Document::STATUS_DRAFT) ? 'Brouillon' : 'Validé';
    }

    public function getStatusCssClass(DocumentInterface $document): string
    {
        return ($document->status == Document::STATUS_DRAFT) ? 'default' : 'success';
    }

    public function getHtmlLabel(DocumentInterface $document)
    {
        $label = $this->getStatusWording($document);
        $classLabel = $this->getStatusCssClass($document);

        return '<span class="label label-' . $classLabel . '">' . $label . '</span>';
    }

    public function isStatus(DocumentInterface $document, string $status): bool
    {
        return $document->status == $status;
    }

    public function isStatusDraft(DocumentInterface $document): bool
    {
        return $this->isStatus($document, Document::STATUS_DRAFT);
    }

    public function isStatusValid(DocumentInterface $document): bool
    {
        return $this->isStatus($document, Document::STATUS_VALID);
    }

    public function getProductsOrders(DocumentInterface $document): array
    {
        $productsOrdersArray = [];
        $ordersArray = $document->orders;
        if ($ordersArray && count($ordersArray)) {
            foreach ($ordersArray as $order) {
                foreach ($order->productOrder as $productOrder) {
                    if ($productOrder->product) {
                        $indexProductOrder = $productOrder->product->order;
                        $newProductOrder = clone $productOrder;

                        if (!isset($productsOrdersArray[$indexProductOrder])) {
                            $productsOrdersArray[$indexProductOrder] = [$newProductOrder];
                        } else {
                            $productOrderMatch = false;
                            foreach ($productsOrdersArray[$indexProductOrder] as &$theProductOrder) {
                                if ($theProductOrder->unit == $productOrder->unit
                                    && $theProductOrder->price == $productOrder->price
                                    && $theProductOrder->invoice_price == $productOrder->invoice_price) {

                                    $theProductOrder->quantity += $productOrder->quantity;
                                    $productOrderMatch = true;
                                }
                            }
                            if (!$productOrderMatch) {
                                $productsOrdersArray[$indexProductOrder][] = $newProductOrder;
                            }
                        }
                    }
                }
            }
        }

        // tri des orderProduct par product.order
        ksort($productsOrdersArray);

        return $productsOrdersArray;
    }

    public function getAliasDirectoryBase(DocumentInterface $document): string
    {
        return '@app/web/pdf/' . $document->id_producer . '/';
    }

    public function getFilename(DocumentInterface $document): string
    {
        $filename = $document->getType() . '-';

        if($this->isStatusValid($document)) {
            $filename .= $document->reference;
        }
        elseif($this->isStatusDraft($document)) {
            $filename .= 'Brouillon-'.$document->id;
        }

        $filename .= '.pdf';

        return $filename;
    }

    public function getFilenameComplete(DocumentInterface $document): string
    {
        return \Yii::getAlias($this->getAliasDirectoryBase($document) . $this->getFilename($document));
    }

    public function isInvoicePrice(DocumentInterface $document): bool
    {
        return $this->getClass() == 'Invoice' || $this->getClass() == 'DeliveryNote';
    }

    public function isTaxCalculationMethodSumOfRoundings(DocumentInterface $document): bool
    {
        return $document->tax_calculation_method == self::TAX_CALCULATION_METHOD_SUM_OF_ROUNDINGS;
    }

    public function isTaxCalculationMethodRoundingOfTheSum(DocumentInterface $document): bool
    {
        return $document->tax_calculation_method == self::TAX_CALCULATION_METHOD_ROUNDING_OF_THE_SUM;
    }
}