<?php

namespace common\logic\Order\Order;

use common\helpers\GlobalParam;
use common\helpers\Price;
use common\logic\BaseService;
use common\logic\BuilderInterface;
use common\logic\Config\TaxRate\TaxRate;
use common\logic\Distribution\Distribution\Distribution;
use common\logic\Distribution\Distribution\DistributionRepository;
use common\logic\Document\Document\Document;
use common\logic\Order\OrderStatusHistory\OrderStatusHistoryBuilder;
use common\logic\Order\ProductOrder\ProductOrder;
use common\logic\Order\ProductOrder\ProductOrderBuilder;
use common\logic\Order\ProductOrder\ProductOrderSolver;
use common\logic\PointSale\PointSale\PointSaleBuilder;
use common\logic\PointSale\PointSale\PointSaleRepository;
use common\logic\PointSale\PointSale\UserPointSaleRepository;
use common\logic\Producer\Producer\Producer;
use common\logic\Producer\Producer\ProducerRepository;
use common\logic\Subscription\Subscription\Subscription;
use common\logic\Subscription\Subscription\SubscriptionBuilder;
use common\logic\Subscription\Subscription\SubscriptionRepository;
use common\logic\Subscription\Subscription\SubscriptionSolver;
use common\logic\User\CreditHistory\CreditHistory;
use common\logic\User\CreditHistory\CreditHistoryBuilder;
use common\logic\User\CreditHistory\CreditHistoryRepository;
use common\logic\User\User\User;
use common\logic\User\User\UserSolver;
use common\logic\User\UserProducer\UserProducerRepository;
use yii\web\NotFoundHttpException;

class OrderBuilder extends BaseService implements BuilderInterface
{
    protected UserSolver $userSolver;
    protected OrderSolver $orderSolver;
    protected CreditHistoryRepository $creditHistoryRepository;
    protected ProducerRepository $producerRepository;
    protected CreditHistoryBuilder $creditHistoryBuilder;
    protected ProductOrderBuilder $productOrderBuilder;
    protected OrderStatusHistoryBuilder $orderStatusHistoryBuilder;
    protected OrderRepository $orderRepository;
    protected DistributionRepository $distributionRepository;
    protected SubscriptionBuilder $subscriptionBuilder;
    protected PointSaleRepository $pointSaleRepository;
    protected UserProducerRepository $userProducerRepository;
    protected UserPointSaleRepository $userPointSaleRepository;
    protected PointSaleBuilder $pointSaleBuilder;
    protected ProductOrderSolver $productOrderSolver;
    protected SubscriptionRepository $subscriptionRepository;
    protected SubscriptionSolver $subscriptionSolver;

    public function __construct()
    {
        $this->userSolver = $this->loadService(UserSolver::class);
        $this->orderSolver = $this->loadService(OrderSolver::class);
        $this->creditHistoryRepository = $this->loadService(CreditHistoryRepository::class);
        $this->producerRepository = $this->loadService(ProducerRepository::class);
        $this->creditHistoryBuilder = $this->loadService(CreditHistoryBuilder::class);
        $this->productOrderBuilder = $this->loadService(ProductOrderBuilder::class);
        $this->orderStatusHistoryBuilder = $this->loadService(OrderStatusHistoryBuilder::class);
        $this->orderRepository = $this->loadService(OrderRepository::class);
        $this->distributionRepository = $this->loadService(DistributionRepository::class);
        $this->subscriptionBuilder = $this->loadService(SubscriptionBuilder::class);
        $this->pointSaleRepository = $this->loadService(PointSaleRepository::class);
        $this->userProducerRepository = $this->loadService(UserProducerRepository::class);
        $this->userPointSaleRepository = $this->loadService(UserPointSaleRepository::class);
        $this->pointSaleBuilder = $this->loadService(PointSaleBuilder::class);
        $this->productOrderSolver = $this->loadService(ProductOrderSolver::class);
        $this->subscriptionRepository = $this->loadService(SubscriptionRepository::class);
        $this->subscriptionSolver = $this->loadService(SubscriptionSolver::class);
    }

    public function instanciate(Distribution $distribution): Order
    {
        $order = new Order();
        $order->populateDistribution($distribution);
        $order->date = date('Y-m-d H:i:s');
        $order->status = 'tmp-order';

        return $order;
    }

    public function create(Distribution $distribution): Order
    {
        $order = $this->instanciate($distribution);

        $this->addUserPointSale($order);
        $this->initCommentPointSale($order);
        $this->generateReference($order);

        $order->save();

        return $order;
    }

    /**
     * Ajoute l'abonnement' pour une date donnée.
     */
    // add
    public function createFromSubscription(Subscription $subscription, string $date, $force = false)
    {
        $now = date('Y-m-d');
        $distributionDate = date('Y-m-d', strtotime($date));
        $distribution = $this->distributionRepository->getOneByDate($distributionDate);

        if ($distribution
            && $distribution->active
            && ($distributionDate > $now || $force)
            && count($subscription->productSubscription)
            && $subscription->pointSale) {

            $order = $this->create($distribution);
            $this->updateFromSubscription($order, $subscription);
        }
    }

    /**
     * Ajoute les commandes pour une date donnée à partir des abonnements.
     */
    // addAll
    public function createAllFromSubscriptions(string $date, $force = false): void
    {
        $distribution = $this->distributionRepository->getOneByDate(date('Y-m-d', strtotime($date)));

        if ($distribution) {
            $orderArray = $this->orderRepository->getByDistribution($distribution);
            $subscriptionArray = $this->subscriptionRepository->getByDate($date);
            foreach ($subscriptionArray as $subscription) {
                if (!$this->subscriptionSolver->hasOrderAlreadyExist($subscription, $orderArray)) {
                    $this->createFromSubscription($subscription, $date, $force);
                }
            }
        }
    }

    public function updateIncomingDistributions(Subscription $subscription, $update = false): void
    {
        $matchedDistributionsArray = $this->distributionRepository->searchMatchedIncomingDistributions($subscription);

        if ($update) {
            $this->deleteOrdersIncomingDistributions($subscription);
        }

        if (count($matchedDistributionsArray)) {
            foreach ($matchedDistributionsArray as $distribution) {
                $this->createFromSubscription($subscription, $distribution->date);
            }
        }
    }

    public function deleteOrdersIncomingDistributions(Subscription $subscription, $deleteAfterDateEnd = false)
    {
        $dateStart = $subscription->date_begin;
        $comparatorDateStart = '>=';
        if ($deleteAfterDateEnd) {
            $dateStart = $subscription->date_end;
            $comparatorDateStart = '>';
        }

        $params = [
            ':id_producer' => GlobalParam::getCurrentProducerId(),
            ':date_today' => date('Y-m-d'),
            ':date_start' => $dateStart,
            ':id_subscription' => $subscription->id
        ];

        $orderDeadline = $this->producerRepository->getConfig('order_deadline');
        $hour = date('G');

        if ($hour >= $orderDeadline) {
            $conditionDistributionDate = 'distribution.date > :date_today';
        } else {
            $conditionDistributionDate = 'distribution.date >= :date_today';
        }

        $orders = Order::find()
            ->joinWith('distribution')
            ->where('distribution.id_producer = :id_producer')
            ->andWhere($conditionDistributionDate)
            ->andWhere('distribution.date ' . $comparatorDateStart . ' :date_start')
            ->andWhere('order.id_subscription = :id_subscription');

        $orders->params($params);

        $ordersArray = $orders->all();
        $configCredit = $this->producerRepository->getConfig('credit');

        $countOrdersDeleted = 0;
        if ($ordersArray && count($ordersArray)) {
            foreach ($ordersArray as $order) {

                $theOrder = $this->orderRepository->getOneById($order->id);

                // remboursement de la commande
                if ($theOrder->id_user && $theOrder->getAmount(Order::AMOUNT_PAID) && $configCredit) {

                    $this->creditHistoryBuilder->create(
                        CreditHistory::TYPE_REFUND,
                        $theOrder->getAmount(Order::AMOUNT_PAID),
                        $theOrder->distribution->producer,
                        $theOrder->user,
                        GlobalParam::getCurrentUser()
                    );
                }

                $this->delete($order, true);

                $countOrdersDeleted ++;
            }
        }

        return $countOrdersDeleted;
    }

    public function addProductOrdersFromSubscription(Order $order, Subscription $subscription): void
    {
        $productsAdd = false;
        $user = $subscription->user ?? null;
        foreach ($subscription->productSubscription as $productSubscription) {

            $this->productOrderBuilder->create(
                $order,
                $productSubscription->product,
                $productSubscription->quantity,
                $productSubscription->product->getPrice([
                    'user' => $user,
                    'point_sale' => $subscription->pointSale,
                    'quantity' => $productSubscription->quantity
                ])
            );
            $productsAdd = true;
        }

        if(!$productsAdd) {
            $order->delete();
        }
    }

    public function updateFromSubscription(Order $order, Subscription $subscription): void
    {
        $this->initBaseFromSubscription($order, $subscription);
        $this->initAutoPaymentFromSubscription($order, $subscription);
        $this->addProductOrdersFromSubscription($order, $subscription);
    }

    public function initBaseFromSubscription(Order $order, Subscription $subscription): void
    {
        $order->origin = Order::ORIGIN_AUTO;
        $order->populatePointSale($subscription->pointSale);
        $order->populateSubscription($subscription);

        if (strlen($subscription->comment)) {
            $order->comment = $subscription->comment;
        }

        if (strlen($subscription->username)) {
            $order->username = $subscription->username;
            $order->id_user = 0;
        } else {
            $order->populateUser($subscription->user);
        }
    }

    public function initAutoPaymentFromSubscription(Order $order, Subscription $subscription): void
    {
        $pointSale = $subscription->pointSale;
        if($pointSale) {
            $creditFunctioning = $this->producerRepository->getPointSaleCreditFunctioning($pointSale);

            $order->auto_payment = 0;
            if ($subscription->auto_payment == Subscription::AUTO_PAYMENT_DEDUCTED) {
                if ($order->id_user && $this->producerRepository->getConfig('credit') && $pointSale->credit) {
                    if ($creditFunctioning == Producer::CREDIT_FUNCTIONING_OPTIONAL) {
                        $order->auto_payment = 0;
                    } elseif ($creditFunctioning == Producer::CREDIT_FUNCTIONING_MANDATORY) {
                        $order->auto_payment = 1;
                    } elseif ($creditFunctioning == Producer::CREDIT_FUNCTIONING_USER) {

                        $userProducer = $this->userProducerRepository->getOne($user, $distribution->producer);

                        if ($userProducer) {
                            $order->auto_payment = $userProducer->credit_active;
                        }
                    }
                }
            } elseif ($subscription->auto_payment == Subscription::AUTO_PAYMENT_YES) {
                $order->auto_payment = 1;
            } elseif ($subscription->auto_payment == Subscription::AUTO_PAYMENT_NO) {
                $order->auto_payment = 0;
            }
        }

        $order->tiller_synchronization = $order->auto_payment;
    }

    public function addUserPointSale(Order $order)
    {
        if($order->user) {
            $this->pointSaleBuilder->addUser($order->user, $order->pointSale);
        }
    }

    public function initCommentPointSale(Order $order): void
    {
        if($order->user && $order->pointSale) {
            $userPointSale = $this->userPointSaleRepository->getOne($order->user, $order->pointSale);

            if ($userPointSale && strlen($userPointSale->comment)) {
                $order->comment_point_sale = $userPointSale->comment;
            }
        }
    }

    /**
     * Initialise le montant total, le montant déjà payé et le poids de la commande.
     */
    public function init(Order $order, $taxCalculationMethod = Document::TAX_CALCULATION_METHOD_DEFAULT)
    {
        $this->initAmount($order, $taxCalculationMethod);
        $this->initPaidAmount($order);

        return $this;
    }

    /**
     * Initialise le montant de la commande.
     */
    public function initAmount(Order $order, $taxCalculationMethod = Document::TAX_CALCULATION_METHOD_DEFAULT)
    {
        $order->amount = 0;
        $order->amount_with_tax = 0;
        $order->amount_vat = [];
        $order->invoice_amount = 0;
        $order->invoice_amount_with_tax = 0;
        $order->invoice_amount_vat = [];
        $order->weight = 0;

        if (isset($order->productOrder)) {
            foreach ($order->productOrder as $productOrder) {
                $this->addAmount($order, Order::AMOUNT_TOTAL, $productOrder, $taxCalculationMethod);
                $this->addAmount($order, Order::INVOICE_AMOUNT_TOTAL, $productOrder, $taxCalculationMethod);
                $this->addWeight($order, $productOrder);
            }
        }
    }

    public function addWeight(Order $order, ProductOrder $productOrder): void
    {
        if ($productOrder->unit == 'piece') {
            if (isset($productOrder->product)) {
                $order->weight += ($productOrder->quantity * $productOrder->product->weight) / 1000;
            }
        } else {
            $order->weight += $productOrder->quantity;
        }
    }

    public function addAmount(Order $order, string $typeTotal, ProductOrder $productOrder, string $taxCalculationMethod): void
    {
        $fieldNameAmount = $this->orderSolver->getFieldNameAmount($typeTotal);
        $fieldNameAmountWithTax = $this->orderSolver->getFieldNameAmount($typeTotal, 'with_tax');
        $price = $this->productOrderSolver->getPriceByTypeTotal($productOrder, $typeTotal);
        $order->$fieldNameAmount += $price * $productOrder->quantity;
        $order->$fieldNameAmountWithTax += Price::getPriceWithTax(
                $price,
                $productOrder->taxRate->value,
                $taxCalculationMethod
            ) * $productOrder->quantity;
        $this->addVat($order, $typeTotal, $price * $productOrder->quantity, $productOrder->taxRate, $taxCalculationMethod);
    }

    public function addVat(
        Order $order,
        string $typeTotal,
        float $priceTotalWithoutTax,
        TaxRate $taxRate,
        string $taxCalculationMethod): void
    {
        $fieldName = $this->orderSolver->getFieldNameAmount($typeTotal, 'vat');

        if (!isset($this->$fieldName[$taxRate->id])) {
            $this->$fieldName[$taxRate->id] = 0;
        }

        $this->$fieldName[$taxRate->id] += Price::getVat($priceTotalWithoutTax, $taxRate->value, $taxCalculationMethod);
    }

    /**
     * Initialise le montant payé de la commande et le retourne.
     */
    public function initPaidAmount(Order $order): void
    {
        if (isset($order->creditHistory)) {
            $history = $order->creditHistory;
        } else {
            $history = $this->creditHistoryRepository->getByOrder($order);
        }

        $order->paid_amount = 0;

        if (count($history)) {
            foreach ($history as $ch) {
                if ($ch->type == CreditHistory::TYPE_PAYMENT) {
                    $order->paid_amount += $ch->amount;
                } elseif ($ch->type == CreditHistory::TYPE_REFUND) {
                    $order->paid_amount -= $ch->amount;
                }
            }
        }
    }

    public function delete(Order $order, $force = false)
    {
        // remboursement si l'utilisateur a payé pour cette commande
        $amountPaid = $this->orderSolver->getAmount(Order::AMOUNT_PAID);

        if ($amountPaid > 0.01) {
            $this->creditHistoryBuilder->create(
                CreditHistory::TYPE_REFUND,
                $amountPaid,
                GlobalParam::getCurrentProducer(),
                $order->user,
                $this->userSolver->getCurrent(),
                $order
            );
        }

        // delete
        if ($this->producerRepository->getConfig('option_behavior_cancel_order') == Producer::BEHAVIOR_DELETE_ORDER_DELETE ||
            ($this->producerRepository->getConfig('option_behavior_cancel_order') == Producer::BEHAVIOR_DELETE_ORDER_STATUS && strlen($order->date_delete))
            || $force) {

            $this->productOrderBuilder->deleteByOrder($order);

            return $order->delete();
        }
        // status 'delete'
        elseif ($this->producerRepository->getConfig('option_behavior_cancel_order') == Producer::BEHAVIOR_DELETE_ORDER_STATUS) {
            $order->date_delete = date('Y-m-d H:i:s');

            return $order->save();
        }
    }

    /**
     * Ajuste le crédit pour que la commande soit payée.
     */
    public function processCredit(Order $order): void
    {
        if ($order->id_user) {
            $paymentStatus = $this->orderSolver->getPaymentStatus($order);

            if ($paymentStatus == Order::PAYMENT_PAID) {
                return;
            } elseif ($paymentStatus == Order::PAYMENT_SURPLUS) {
                $type = CreditHistory::TYPE_REFUND;
                $amount = $this->orderSolver->getAmount($order, Order::AMOUNT_SURPLUS);
            } elseif ($paymentStatus == Order::PAYMENT_UNPAID) {
                $type = CreditHistory::TYPE_PAYMENT;
                $amount = $this->orderSolver->getAmount($order, Order::AMOUNT_REMAINING);
            }

            $this->creditHistoryBuilder->create(
                $type,
                $amount,
                GlobalParam::getCurrentProducer(),
                $order->user,
                $this->userSolver->getCurrent(),
                $order
            );
        }
    }

    // setTillerSynchronization
    public function updateTillerSynchronization(Order $order)
    {
        $paymentStatus = $order->getPaymentStatus();

        if ($paymentStatus == Order::PAYMENT_PAID) {
            $order->tiller_synchronization = 1;
        } else {
            $order->tiller_synchronization = 0;
        }

        $order->save();

        return $order;
    }

    /**
     * Changement de statut d'une commande.
     */
    public function changeOrderStatus(Order $order, string $newStatus, string $origin): void
    {
        $orderStatusArray = GlobalParam::get('orderStatus');
        $userCurrent = $this->userSolver->getCurrent();

        switch ($newStatus) {
            case 'new-order' :
                $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                $order->status = $newStatus;
                $order->save();
                break;
            case 'waiting-paiement-on-delivery':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'waiting-paiement-by-credit':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'paid-by-credit':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'waiting-delevery' :
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'delivered':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'refunded':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'cancel':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            default:
                throw new NotFoundHttpException('Statut de commande inconnu.');
        }
    }

    // initReference
    public function generateReference(Order $order)
    {
        $idProducer = GlobalParam::getCurrentProducerId();
        $producer = $this->producerRepository->getOneById($idProducer);

        if (!$order->reference && $producer->option_order_reference_type == Producer::ORDER_REFERENCE_TYPE_YEARLY) {
            $lastOrder = $this->orderRepository->getOneLastOfYear();

            if ($lastOrder && $lastOrder->reference && strlen($lastOrder->reference) > 0) {
                $pattern = '#A([0-9]+)C([0-9]+)#';
                preg_match($pattern, $lastOrder->reference, $matches, PREG_OFFSET_CAPTURE);
                $sizeNumReference = strlen($matches[2][0]);
                $numReference = ((int)$matches[2][0]) + 1;
                $numReference = str_pad($numReference, $sizeNumReference, '0', STR_PAD_LEFT);

                $order->reference = 'A' . $matches[1][0] . 'C' . $numReference;
            } else {
                $order->reference = 'A' . date('y') . 'C0001';
            }

            $order->save();
        }
    }

    // initInvoicePrices
    public function updateInvoicePrices(Order $order, $params = []): void
    {
        foreach ($order->productOrder as $productOrder) {
            if ($productOrder->product) {
                $this->productOrderBuilder->updateInvoicePrice($productOrder, $params);
            }
        }
    }

}